import dataclasses
import functools
import pathlib

import aiida
import aiida.common
import aiida.orm
import pytest

from aiida_icon.calculations import IconCalculation

pytest_plugins = ["aiida.tools.pytest_fixtures"]


@dataclasses.dataclass
class ParserCase:
    datapath: pathlib.Path
    exit_code: int
    required_output_links: list[str]
    disallowed_output_links: list[str]
    finish_status_value: str


PARSER_CASES = {
    "simple_icon_run": ("simple_icon_run", 0, ["finish_status"], [], "OK"),
    "restarts_present": (
        "restarts_present",
        0,
        ["finish_status", "latest_restart_file", "all_restart_files"],
        [],
        "RESTART",
    ),
    "restarts_missing": (
        "restarts_missing",
        304,
        ["finish_status"],
        ["latest_restart_file", "all_restart_files"],
        "RESTART",
    ),
}


@pytest.fixture(params=["simple_icon_run", "restarts_present", "restarts_missing"])
def case_name(request):
    return request.param


@pytest.fixture
def parser_case(case_name):
    case_name, *args = PARSER_CASES[case_name]
    return ParserCase(pathlib.Path(__file__).parent.absolute() / "data" / case_name, *args)


@dataclasses.dataclass
class BuildInputs:
    fake_icon: aiida.orm.CalcJobNode

    def __setattr__(self, name: str, value: aiida.orm.Data) -> None:
        if name == "fake_icon":
            super().__setattr__(name, value)
        else:
            self.fake_icon.base.links.add_incoming(
                source=value,
                link_type=aiida.common.LinkType.INPUT_CALC,
                link_label=name,
            )


@dataclasses.dataclass
class BuildOutputs:
    fake_icon: aiida.orm.CalcJobNode

    def __setattr__(self, name: str, value: aiida.orm.Data) -> None:
        if name == "fake_icon":
            super().__setattr__(name, value)
        else:
            value.base.links.add_incoming(
                link_type=aiida.common.LinkType.CREATE,
                source=self.fake_icon,
                link_label=name,
            )


class FakeIconBuilder:
    node: aiida.orm.CalcJobNode

    def __init__(self, computer: aiida.orm.Computer):
        self.node = aiida.orm.CalcJobNode(computer=computer, process_type="aiida.calculations:icon.icon")

    @property
    def inputs(self) -> BuildInputs:
        return BuildInputs(self.node)

    @property
    def outputs(self) -> BuildOutputs:
        return BuildOutputs(self.node)

    def build(self) -> aiida.orm.CalcJobNode:
        self.node.store_all()
        return self.node


@pytest.fixture
def icon_result(parser_case, aiida_computer_local):
    """Mockup a finished calculation for a given set of inputs and outputs."""
    datapath = parser_case.datapath
    computer = aiida_computer_local()
    make_remote = functools.partial(aiida.orm.RemoteData, computer=computer)
    builder = FakeIconBuilder(computer=computer)
    builder.inputs.master_namelist = aiida.orm.SinglefileData(datapath / "inputs" / "icon_master.namelist")
    builder.inputs.model_namelist = aiida.orm.SinglefileData(datapath / "inputs" / "model.namelist")
    builder.inputs.dynamics_grid_file = make_remote(
        remote_path=str(datapath.absolute() / "inputs" / "icon_grid_simple.nc")
    )
    builder.inputs.ecrad_data = make_remote(remote_path=str(datapath.absolute() / "inputs" / "ecrad_data"))
    builder.inputs.rrtmg_sw = make_remote(remote_path=str(datapath.absolute() / "inputs" / "rrtmg_sw.nc"))
    builder.inputs.cloud_opt_props = make_remote(
        remote_path=str(datapath.absolute() / "inputs" / "ECHAM6_CldOptProps.nc")
    )
    builder.inputs.dmin_wetgrowth_lookup = make_remote(
        remote_path=str(datapath.absolute() / "inputs" / "dmin_wetgrowth_lookup.nc")
    )
    node = builder.build()
    builder.outputs.remote_folder = make_remote(str(datapath.absolute() / "outputs")).store()

    retrieved = aiida.orm.FolderData()
    retrieved_files = [
        "_aiidasubmit.sh",
        "_scheduler-stderr.txt",
        "_scheduler-stdout.txt",
        "finish.status",
    ]
    for filename in retrieved_files:
        retrieved.put_object_from_file(str(datapath.absolute() / "outputs" / filename), filename)
    builder.outputs.retrieved = retrieved.store()

    return node


@pytest.fixture
def icon_calc(aiida_computer_local, aiida_code_installed):
    """Create an IconCalculation which is ready to call .prepare_for_submission()."""
    code = aiida_code_installed(default_calc_job_plugin="icon.icon", computer=aiida_computer_local())
    datapath = pathlib.Path(__file__).parent.absolute() / "data" / "simple_icon_run"
    builder = code.get_builder()
    make_remote = functools.partial(aiida.orm.RemoteData, computer=code.computer)
    builder.master_namelist = aiida.orm.SinglefileData(datapath / "inputs" / "icon_master.namelist")
    builder.model_namelist = aiida.orm.SinglefileData(datapath / "inputs" / "model.namelist")
    builder.dynamics_grid_file = make_remote(remote_path=str(datapath.absolute() / "inputs" / "icon_grid_simple.nc"))
    builder.ecrad_data = make_remote(remote_path=str(datapath.absolute() / "inputs" / "ecrad_data"))
    builder.rrtmg_sw = make_remote(remote_path=str(datapath.absolute() / "inputs" / "rrtmg_sw.nc"))
    builder.cloud_opt_props = make_remote(remote_path=str(datapath.absolute() / "inputs" / "ECHAM6_CldOptProps.nc"))
    builder.dmin_wetgrowth_lookup = make_remote(
        remote_path=str(datapath.absolute() / "inputs" / "dmin_wetgrowth_lookup.nc")
    )
    return IconCalculation(dict(builder))



from __future__ import annotations

import dataclasses
import enum
import pathlib
import re
import typing

import f90nml
from aiida import engine, orm
from aiida.common import datastructures, folders
from aiida.parsers import parser

from aiida_icon import exceptions
from aiida_icon.iconutils import masternml, modelnml

if typing.TYPE_CHECKING:
    from aiida.engine.processes.calcjobs import calcjob


class IconCalculation(engine.CalcJob):
    """AiiDA calculation to run ICON."""

    @classmethod
    def define(cls, spec: calcjob.CalcJobProcessSpec) -> None:  # type: ignore[override] # forced by aiida-core
        super().define(spec)
        spec.input("master_namelist", valid_type=orm.SinglefileData)
        spec.input("model_namelist", valid_type=orm.SinglefileData)
        spec.input("restart_file", valid_type=orm.RemoteData, required=False)
        spec.input("wrapper_script", valid_type=orm.SinglefileData, required=False)
        spec.input(
            "dynamics_grid_file",
            valid_type=orm.RemoteData,
        )
        spec.input("ecrad_data", valid_type=orm.RemoteData)
        spec.input("cloud_opt_props", valid_type=orm.RemoteData)
        spec.input("dmin_wetgrowth_lookup", valid_type=orm.RemoteData)
        spec.input("rrtmg_sw", valid_type=orm.RemoteData)
        spec.output("latest_restart_file")
        spec.output_namespace("all_restart_files", dynamic=True)
        spec.output("finish_status")
        options = spec.inputs["metadata"]["options"]  # type: ignore[index] # guaranteed correct by aiida-core
        options["resources"].default = {  # type: ignore[index] # guaranteed correct by aiida-core
            "num_machines": 10,
            "num_mpiprocs_per_machine": 1,
            "num_cores_per_mpiproc": 2,
        }
        options["withmpi"].default = True  # type: ignore[index] # guaranteed correct by aiida-core
        options["parser_name"].default = "icon.icon"  # type: ignore[index] # guaranteed correct by aiida-core
        spec.exit_code(
            300,
            "ERROR_MISSING_OUTPUT_FILES",
            message="ICON did not create a restart file or directory.",
        )
        spec.exit_code(
            301,
            "ERROR_READING_STATUS_FILE",
            message="Could not read the finish.status file.",
        )
        # deprecated, replaced by 304: PARTIALLY_PARSED + log messages
        spec.exit_code(
            302,
            "FAILED_CHECK_STATUS",
            message="The final status was not 'OK or RESTART', check the finish_status output.",
        )
        # deprecated, replaced by 304: PARTIALLY_PARSED + log messages
        spec.exit_code(
            303,
            "UNSUPPORTED_FEATURE",
            message="Could not fully parse due to an unsupported feature, please check the log.",
        )
        spec.exit_code(
            304,
            "PARTIALLY_PARSED",
            message="Some outputs might be missing, check the log for explanations.",
        )
        # deprecated, replaced by 304: PARTIALLY_PARSED + log messages
        spec.exit_code(
            310,
            "ERROR_MISSING_RESTART_FILES",
            message="ICON was expected to produce a restart file but did not.",
        )

    def prepare_for_submission(self, folder: folders.Folder) -> datastructures.CalcInfo:
        model_namelist_data = f90nml.reads(self.inputs.model_namelist.get_content())
        master_namelist_data = f90nml.reads(self.inputs.master_namelist.get_content())

        for output_folder in modelnml.read_output_stream_paths(model_namelist_data):
            folder.get_subfolder(output_folder, create=True)

        codeinfo = datastructures.CodeInfo()
        codeinfo.code_uuid = self.inputs.code.uuid

        calcinfo = datastructures.CalcInfo()
        calcinfo.codes_info = [codeinfo]
        calcinfo.remote_symlink_list = [
            (
                self.inputs.code.computer.uuid,
                self.inputs.dynamics_grid_file.get_remote_path(),
                model_namelist_data["grid_nml"]["dynamics_grid_filename"].strip(),
            ),
            (
                self.inputs.code.computer.uuid,
                self.inputs.ecrad_data.get_remote_path(),
                model_namelist_data["radiation_nml"]["ecrad_data_path"].strip(),
            ),
            (
                self.inputs.code.computer.uuid,
                self.inputs.rrtmg_sw.get_remote_path(),
                "rrtmg_sw.nc",
            ),
        ]
        if "restart_file" in self.inputs:
            calcinfo.remote_symlink_list.append(
                (
                    self.inputs.code.computer.uuid,
                    self.inputs.restart_file.get_remote_path(),
                    modelnml.read_latest_restart_file_link_name(model_namelist_data),
                )
            )
        calcinfo.remote_copy_list = [
            (
                self.inputs.code.computer.uuid,
                self.inputs.cloud_opt_props.get_remote_path(),
                "ECHAM6_CldOptProps.nc",
            ),
            (
                self.inputs.code.computer.uuid,
                self.inputs.dmin_wetgrowth_lookup.get_remote_path(),
                "dmin_wetgrowth_lookup.nc",
            ),
        ]
        calcinfo.local_copy_list = [
            (
                self.inputs.master_namelist.uuid,
                self.inputs.master_namelist.filename,
                "icon_master.namelist",
            ),
            (
                self.inputs.model_namelist.uuid,
                self.inputs.model_namelist.filename,
                master_namelist_data["master_model_nml"]["model_namelist_filename"].strip(),
            ),
        ]

        if "wrapper_script" in self.inputs:
            calcinfo.local_copy_list.append(
                (
                    self.inputs.wrapper_script.uuid,
                    self.inputs.wrapper_script.filename,
                    "run_icon.sh",
                )
            )
        calcinfo.retrieve_list = [
            "finish.status",
            "nml.atmo.log",
            "output_schedule.txt",
        ]
        return calcinfo


class FinishStatus(enum.Enum):
    OK = enum.auto()
    RESTART = enum.auto()
    UNEXPECTED = enum.auto()
    ERR_READING_STATUS = enum.auto()
    ERR_MISSING_STATUS = enum.auto()


class RestartStatus(enum.Enum):
    OK = enum.auto()
    MISSING = enum.auto()
    ERROR = enum.auto()


@dataclasses.dataclass
class FinishStatusResult:
    status: FinishStatus
    message: orm.Str | None = None


@dataclasses.dataclass
class RestartResult:
    status: RestartStatus
    all_restarts: dict[str, orm.RemoteData] = dataclasses.field(default_factory=dict)
    latest_restart: orm.RemoteData | None = None


class IconParser(parser.Parser):
    """Parser for raw Icon calculations."""

    def parse(self, **kwargs):  # noqa: ARG002  # kwargs must be there for superclass compatibility
        finish_status = self.parse_finish_status()
        if finish_status.message:
            self.out("finish_status", finish_status.message)

        restart_indicated = finish_status is FinishStatus.RESTART or masternml.read_lrestart_write_last(
            self.node.inputs.master_namelist
        )
        restarts = self.parse_restart_files(restart_indicated=restart_indicated)
        if restarts.all_restarts:
            self.out("all_restart_files", restarts.all_restarts)
        if restarts.latest_restart:
            self.out("latest_restart_file", restarts.latest_restart)

        match finish_status.status:
            case FinishStatus.OK:
                pass
            case FinishStatus.RESTART:
                if restarts.status is not RestartStatus.OK:
                    return self.exit_codes.PARTIALLY_PARSED
            case FinishStatus.UNEXPECTED:
                return self.exit_codes.PARTIALLY_PARSED
            case FinishStatus.ERR_READING_STATUS:
                return self.exit_codes.ERROR_READING_STATUS_FILE
            case FinishStatus.ERR_MISSING_STATUS:
                return self.exit_codes.ERROR_MISSING_OUTPUT_FILES

        return engine.ExitCode(0)

    def parse_finish_status(self) -> FinishStatusResult:
        result = FinishStatusResult(status=FinishStatus.ERR_MISSING_STATUS, message=None)
        if "finish.status" in self.retrieved.list_object_names():
            try:
                with self.retrieved.open("finish.status", "r") as status_file:
                    out_status = status_file.read().strip()
                    result.message = orm.Str(out_status)
                    match out_status:
                        case "OK":
                            result.status = FinishStatus.OK
                        case "RESTART":
                            result.status = FinishStatus.RESTART
                        case _:
                            result.status = FinishStatus.UNEXPECTED
                            self.logger.info("The 'finish.status' file contained an unexpected value.")
            except OSError as err:
                result.status = FinishStatus.ERR_READING_STATUS
                self.logger.warning(str(err))
                self.logger.warning("The 'finish.status' file was found but could not be read.")
        else:
            self.logger.warning("The 'finish.status' file could not be found in the output.")

        return result

    def parse_restart_files(self, *, restart_indicated: bool) -> RestartResult:
        remote_folder = self.node.outputs.remote_folder
        remote_path = pathlib.Path(remote_folder.get_remote_path())

        files = remote_folder.listdir()
        result = RestartResult(status=RestartStatus.MISSING)
        try:
            all_restarts_pattern = modelnml.read_restart_file_pattern(self.node.inputs.model_namelist)
            latest_restart_name = modelnml.read_latest_restart_file_link_name(self.node.inputs.model_namelist)
        except exceptions.SinglefileRestartNotImplementedError:
            self.logger.info("Can not parse restart file names, singlefile mode is not supported.")
            if restart_indicated:
                result.status = RestartStatus.ERROR

        for file_name in files:
            if restart_match := re.match(all_restarts_pattern, file_name):
                result.all_restarts[f"restart_{restart_match['timestamp']}"] = orm.RemoteData(
                    computer=self.node.computer,
                    remote_path=str(remote_path / file_name),
                )
            if file_name == latest_restart_name:
                result.latest_restart = orm.RemoteData(
                    computer=self.node.computer,
                    remote_path=str(remote_path / file_name),
                )

        if result.all_restarts and result.latest_restart:
            result.status = RestartStatus.OK
        else:
            self.logger.info("Could not find a valid set of restart files.")

        return result



import pathlib

import f90nml

from aiida_icon import exceptions
from aiida_icon.iconutils import namelists


def read_restart_file_pattern(model_nml: namelists.NMLInput) -> str:
    data = namelists.namelists_data(model_nml)

    restart_write_mode = data.get("io_nml", {}).get("restart_write_mode", "joint procs multifile")
    if "multifile" not in restart_write_mode:
        raise exceptions.SinglefileRestartNotImplementedError

    return r"multifile_restart_atm_(?P<timestamp>\d{8}T\d{6}Z).mfr"


def read_latest_restart_file_link_name(model_nml: namelists.NMLInput) -> str:
    data = namelists.namelists_data(model_nml)

    restart_write_mode = data.get("io_nml", {}).get("restart_write_mode", "joint procs multifile")
    if "multifile" not in restart_write_mode:
        raise exceptions.SinglefileRestartNotImplementedError

    return "multifile_restart_atm.mfr"


def read_output_stream_paths(
    model_nml: namelists.NMLInput,
) -> list[pathlib.Path]:
    data = namelists.namelists_data(model_nml)

    output_data = data["output_nml"]

    # wrap in list if there is only one
    stream_spec_list: list[f90nml.namelist.Namelist] = (
        [output_data] if isinstance(output_data, f90nml.namelist.Namelist) else output_data
    )

    return [_out_stream_path(stream_spec) for stream_spec in stream_spec_list]


def _out_stream_path(out_stream_spec: f90nml.namelist.Namelist) -> pathlib.Path:
    """Replicate ICON logic in forming the filenames to get the output dir."""

    filename_format = out_stream_spec.get("filename_format", "<output_filename>_XXX_YYY")
    output_filename = out_stream_spec.get("output_filename", "")
    return pathlib.Path(filename_format.replace("<output_filename>", output_filename)).parent



import pathlib

import pytest
from aiida.common import folders

from aiida_icon import calculations


def test_prepare_for_calc(icon_calc, tmp_path):
    prepare_path = tmp_path / "test_prepare_simple"
    prepare_path.mkdir()
    sandbox_folder = folders.SandboxFolder(prepare_path.absolute())
    calcinfo = icon_calc.presubmit(sandbox_folder)

    outputs_2d = sandbox_folder.get_subfolder("simple_icon_run_atm_2d", create=False)
    outputs_3d = sandbox_folder.get_subfolder("simple_icon_run_atm_3d_pl", create=False)
    remote_link_names = [triplet[2] for triplet in calcinfo.remote_symlink_list]
    local_copy_names = [triplet[2] for triplet in calcinfo.local_copy_list]

    assert outputs_2d.exists()
    assert outputs_3d.exists()
    assert "simple_icon_run.namelist" in local_copy_names
    assert "icon_grid_simple.nc" in remote_link_names
    assert "./ecrad_data" in remote_link_names


def test_parser_simple(parser_case, icon_result):
    """Parsed output links and exit code should match expectations."""
    parser = calculations.IconParser(icon_result)
    exit_code = parser.parse()

    assert exit_code.status == parser_case.exit_code

    assert all(link in parser.outputs for link in parser_case.required_output_links)
    assert all(link not in parser.outputs for link in parser_case.disallowed_output_links)
    assert parser.outputs.finish_status.value == parser_case.finish_status_value


@pytest.mark.parametrize("case_name", ["restarts_present"])
def test_additional_restart_parsing(case_name, parser_case, icon_result):
    """Check the contents of restarts related parsing outputs."""
    parser = calculations.IconParser(icon_result)
    parser.parse()
    assert pathlib.Path(parser.outputs.latest_restart_file.get_remote_path()).name == "multifile_restart_atm.mfr"
    assert (
        pathlib.Path(parser.outputs.all_restart_files["restart_20000101T030000Z"].get_remote_path()).name
        == "multifile_restart_atm_20000101T030000Z.mfr"
    )



import pathlib
import textwrap

import aiida
from aiida.engine import processes

SCRIPT_DIR = pathlib.Path(__file__).parent.absolute() / "wrapper_scripts"


# TODO(ricoh): allow custom images
def common_alps_setup(builder: processes.ProcessBuilder) -> None:
    """
    Set AiiDA process options for running an aiida_icon.icon calcjob on ALPS.

    Needs to be used with the `wrapper_script` input of the IconCalculation.
    """
    options = builder.metadata.options  # type: ignore[attr-defined]  # builder has a custom setattr
    options.custom_scheduler_commands = "\n".join(
        [
            *options.custom_scheduler_commands.splitlines(),
            "#SBATCH --uenv=icon-wcp/v1:rc4",
        ]
    )
    options.environment_variables = {
        "CUDA_BUFFER_PAGE_IN_THRESHOLD": "0.001",
        "FI_CXI_SAFE_DEVMEM_COPY_THRESHOLD": "0",
        "FI_CXI_RX_MATCH_NODE": "software",
        "FI_MR_CACHE_MONITOR": "disabled",
        "MPICH_GPU_SUPPORT_ENABLED": "1",
        "NVCOMPILER_ACC_DEFER_UPLOADS": "1",
        "NVCOMPILER_TERM": "trace",
        "OMP_NUM_THREADS": "1",
        "ICON_THREADS": "1",
        "OMP_SCHEDULE": "static,1",
        "OMP_DYNAMIC": "false",
        "OMP_STACKSIZE": "200M",
    }
    options.prepend_text = textwrap.dedent(
        """
        chmod 755 run_icon.sh
        """
    )
    options.mpirun_extra_params.append("./run_icon.sh")


def setup_for_todi_cpu(builder: processes.ProcessBuilder) -> None:
    common_alps_setup(builder)
    builder.wrapper_script = aiida.orm.SinglefileData(file=SCRIPT_DIR / "todi_cpu.sh")



import typing

import aiida.orm
import f90nml

NMLInput: typing.TypeAlias = aiida.orm.SinglefileData | f90nml.namelist.Namelist


def namelists_data(
    namelist: NMLInput,
) -> f90nml.namelist.Namelist:
    match namelist:
        case f90nml.namelist.Namelist():
            return namelist
        case aiida.orm.SinglefileData():
            return f90nml.reads(namelist.get_content())
        case _:
            raise ValueError



import functools
import pathlib

import aiida
import aiida.engine
import aiida.orm
from aiida_icon.site_support.cscs.todi import setup_for_todi_cpu

COMPUTER_NAME = "todi"
ICON_CODE_NAME = "icon"
DRY_RUN = False

if __name__ == "__main__":
    thisdir = pathlib.Path(__file__).parent.absolute()
    icon = aiida.orm.load_code(f"{ICON_CODE_NAME}@{COMPUTER_NAME}")
    make_remote_data = functools.partial(aiida.orm.RemoteData, computer=icon.computer)
    builder = icon.get_builder()
    builder.master_namelist = aiida.orm.SinglefileData(file=thisdir / "icon_master.namelist")
    builder.model_namelist = aiida.orm.SinglefileData(file=thisdir / "NAMELIST_exclaim_ape_R02B04")
    builder.dynamics_grid_file = make_remote_data(
        remote_path="/ABSOLUTE_PATH_TO/icon_grid_0013_R02B04_R.nc"
    )  # filename must match the model namelist contents
    builder.ecrad_data = make_remote_data(remote_path="/ABSOLUTE_PATH_TO/ecrad/data")
    builder.rrtmg_sw = make_remote_data(remote_path="/ABSOLUTE_PATH_TO/rrtmg_sw.nc")
    builder.cloud_opt_props = make_remote_data(remote_path="/ABSOLUTE_PATH_TO/ECHAM6_CldOptProps.nc")
    builder.dmin_wetgrowth_lookup = make_remote_data(remote_path="/ABSOLUTE_PATH_TO/dmin_wetgrowth_lookup.nc")
    builder.metadata.options.mpirun_extra_params = [
        "--threads-per-core=1",
        "--distribution=block:block:block",
    ]
    builder.metadata.description = "Icon on Todi through via wrapper script."
    builder.metadata.options.resources = {"num_machines": 1, "num_mpiprocs_per_machine": 288}
    builder.metadata.options.max_memory_kb = 128000000
    builder.metadata.dry_run = DRY_RUN

    setup_for_todi_cpu(builder)

    print(aiida.engine.submit(builder))



import pathlib

import aiida
import aiida.orm
import f90nml
import pytest

from aiida_icon.iconutils import masternml


@pytest.fixture
def empty_options():
    return masternml.options(
        master_options=masternml.MasterOptions(),
        time_control_options=masternml.TimeControlOptions(),
    )


@pytest.fixture
def write_restart_options():
    return masternml.options(
        master_options=masternml.MasterOptions(lrestart=True),
        time_control_options=masternml.TimeControlOptions(
            experiment_start_date="2000-01-01T00:00:00Z",
            experiment_stop_date="2000-01-01T06:00:00Z",
            restart_time_int_val="PT3H",
            checkpoint_time_int_val="PT3H",
        ),
    )


def test_options_empty(empty_options):
    assert empty_options == {"master_nml": {}, "master_time_control_nml": {}}


def test_write_restart_options(write_restart_options):
    assert write_restart_options == {
        "master_nml": {"lrestart": True},
        "master_time_control_nml": {
            "experiment_start_date": "2000-01-01T00:00:00Z",
            "experiment_stop_date": "2000-01-01T06:00:00Z",
            "restart_time_int_val": "PT3H",
            "checkpoint_time_int_val": "PT3H",
        },
    }


def test_modify_master_nml(write_restart_options):
    master_nml = aiida.orm.SinglefileData(
        pathlib.Path(__file__).parent.parent.parent / "examples" / "exclaim_R02B04" / "icon_master.namelist"
    )
    options = aiida.orm.Dict(write_restart_options)
    modified = masternml.modify_master_nml(master_nml, options)

    assert master_nml.is_stored
    assert options.is_stored
    assert modified.is_stored

    data = f90nml.reads(modified.get_content())
    for section, section_data in options.items():
        for key, value in section_data.items():
            assert data[section][key] == value



from __future__ import annotations

import dataclasses
import io
from typing import Any

import aiida.engine
import aiida.orm
import f90nml

from aiida_icon.iconutils import namelists


@dataclasses.dataclass
class OptionsMixin:
    """
    Overrides mixin for dataclasses.

    Add an as_dict method, which ignores keys with value None.
    """

    def as_dict(self) -> dict[str, Any]:
        return {k: v for k, v in dataclasses.asdict(self).items() if v is not None}


@dataclasses.dataclass
class MasterOptions(OptionsMixin):
    lrestart: bool | None = None
    read_restart_namelists: bool | None = None


@dataclasses.dataclass
class TimeControlOptions(OptionsMixin):
    experiment_start_date: str | None = None
    experiment_stop_date: str | None = None
    restart_time_int_val: str | None = None
    checkpoint_time_int_val: str | None = None


def options(master_options: MasterOptions, time_control_options: TimeControlOptions) -> dict[str, dict[str, Any]]:
    return {
        "master_nml": master_options.as_dict(),
        "master_time_control_nml": time_control_options.as_dict(),
    }


@aiida.engine.calcfunction
def modify_master_nml(master_nml: aiida.orm.SinglefileData, options: aiida.orm.Dict) -> aiida.orm.SinglefileData:
    data = f90nml.reads(master_nml.get_content())
    for section in data:
        if section in options:
            data[section] |= options[section]
    string_buffer = io.StringIO()
    f90nml.write(data, string_buffer)
    return aiida.orm.SinglefileData(io.BytesIO(bytes(string_buffer.getvalue(), "utf8")))


def read_lrestart_write_last(master_nml: namelists.NMLInput) -> bool:
    return namelists.namelists_data(master_nml).get("master_nml", {}).get("lrestart_write_last", False)
